# pylint: disable=F,W
# flake8: noqa
from ZODB.DemoStorage import DemoStorage
from ZODB.FileStorage import FileStorage
from ZODB.MappingStorage import MappingStorage

from .attributes.base import BaseAttribute, set_read_memoization
from .attributes.blobs import Blob, BlobAttribute
from .attributes.collections import (
    DictAttribute,
    LargeDictAttribute,
    LargeListAttribute,
    ListAttribute,
    SetAttribute,
    SmallDictAttribute,
    SmallListAttribute,
)
from .attributes.counter import CounterAttribute
from .attributes.files import (
    FileAttribute,
    FileObject,
    FilesGarbageCollector,
    set_files_root_dir,
)
from .attributes.models import (
    ModelAttribute,
    InlineModelAttribute,
    IndexedModelAttribute,
)
from .attributes.simples import (
    BooleanAttribute,
    DateAttribute,
    DateTimeAttribute,
    FloatAttribute,
    IntegerAttribute,
    SimpleAttribute,
    StringAttribute,
    StringUUIDAttribute,
    TimeAttribute,
    UUIDAttribute,
)
from .constants import ASC, DESC
from .databases import Database, connection
from .exceptions import (
    SherafException,
    ObjectNotFoundException,
    ModelObjectNotFoundException,
    SameNameForTableException,
    NotConnectedException,
    InvalidFilterException,
    InvalidIndexException,
    InvalidOrderException,
    QuerySetUnpackException,
    EmptyQuerySetUnpackException,
    PrimaryKeyException,
)
from .models import (
    AutoModel,
    AttributeModel,
    IntAutoModel,
    IntIndexedIntAttributesModel,
    IntIndexedModel,
    IntIndexedNamedAttributesModel,
    IntOrderedNamedAttributesModel,
    Model,
    UUIDIndexedDatedNamedAttributesModel,
    UUIDIndexedModel,
    UUIDIndexedNamedAttributesModel,
)
from .models.attributes import (
    DatedNamedAttributesModel,
    IntAttributesModel,
    NamedAttributesModel,
)
from .models.base import BaseModel
from .models.indexation import (
    BaseIndexedModel,
    IndexedModel,
)
from .models.inline import InlineModel
from .queryset import QuerySet
from .transactions import attempt, commit
from .version import __version__, __version_info__



class ModelLoader(object):
    """
    Loads models from the base in a cache.

    Inherited by most model types
    (``Model[Attribute|List|Set|(Large)Dict]``)
    """

    cache = {}

    def __init__(self, model=None, **kwargs):
        super().__init__(**kwargs)
        self._model = model

    def load_model(self, modelpath):
        # Internal.
        # :param modelpath: namespace of the model to load (eg 'x.y.z')
        # :return: a model class given its path
        if isinstance(modelpath, bytes):
            modelpath = modelpath.decode("utf-8")

        _path = modelpath.split(".")
        _sub_path, _class = ".".join(_path[:-1]), _path[-1]

        _module = __import__(_sub_path, globals(), locals(), [_class], 0)
        # pypy:
        if _module is None:  # pragma: no cover
            raise ImportError(modelpath)

        return getattr(_module, _class)

    def read(self, parent):
        self.check_model(parent)
        return super().read(parent)

    def write(self, parent, value):
        self.check_model(parent)
        return super().write(parent, value)

    @property
    def model(self):
        self.check_model()
        return self._model

    def check_model(self, parent=None):
        if isinstance(self._model, (str, bytes)):
            try:
                self._model = ModelLoader.cache[self._model]
            except KeyError:
                try:
                    ModelLoader.cache[self._model] = self.load_model(self._model)
                except ValueError:
                    raise ImportError

                self._model = ModelLoader.cache[self._model]

        elif parent and not isinstance(self._model, type):
            self._model = type(
                "{}.{}".format(parent.__class__.__name__, self.key(parent)),
                (self._model.__class__,),
                self._model.attributes,
            )



"""Maybe those functions do consume a lot of memory, what about editing
existing dicts instead of creating new ones?"""


class DictConflictException(BaseException):
    pass


def pr_eq(a, b):
    try:
        return a == b
    except ValueError:
        return False


def merge(old, a, b):
    res = {}
    for k in set(a.keys()) | set(b.keys()) | set(old.keys()):
        # value has been deleted in a and b --> ignore
        if k not in a and k not in b:
            continue

        # value not in old
        if k not in old:
            if k not in a:
                # PR is new in b --> keep b
                res[k] = b[k]
                continue

            if k not in b:
                # PR is new in a --> keep a
                res[k] = a[k]
                continue

        # value in old
        else:
            # a unchanged, b changed --> keep b
            if pr_eq(a.get(k), old.get(k)):
                res[k] = b[k]
                continue

            # b unchanged, a changed --> keep a
            if pr_eq(b.get(k), old.get(k)):
                res[k] = a[k]
                continue

        # value equal in a and b --> keep
        if pr_eq(a.get(k), b.get(k)):
            res[k] = a[k]
            continue

        # values are dict --> merge
        if isinstance(a[k], dict) and isinstance(b[k], dict):
            res[k] = merge(old[k], a[k], b[k])
            continue

        raise DictConflictException("Conflict found in key %s" % k)

    return res



class IndexDetails:
    """
    :param attribute: The attribute being indexed
    :type attribute: class BaseAttribute
    :param key: The key the index will use. By default, just the attribute name is used.
    :param unique: If the attribute is unique, and two models have the same value for this
                   attribute, a :class:`~sheraf.exceptions.UniqueIndexException` is raised
                   when trying to write the second one. Automatically set to :class:`True` if
                   primary is :class:`True`.
    :type unique: bool
    :param key: The key the index will use. By default, just the attribute name is used.
    :param values_func: A callable that takes the current attribute value and returns a
                        collection of values to index. Each generated value will be
                        indexed each time this attribute is edited. It may take time if
                        the generated collection is large. By default, the current
                        attribute raw value is used.
    :param search_func: A callable that takes some raw data and returns a collection
                        of values to search in the index. By default, *values_func*
                        is used.
    :param mapping: The mapping object to be used to store the indexed values. OOBTree by
                    default.
    """

    unique = False
    key = None
    values_func = None
    search_func = None
    attribute = None
    mapping = None
    primary = False

    def __init__(
        self, attribute, unique, key, values_func, search_func, mapping, primary
    ):
        def default_values_func(value):
            return {value}

        self.attribute = attribute
        self.unique = unique or primary
        self.key = key
        self.values_func = values_func or default_values_func
        self.search_func = search_func or self.values_func
        self.mapping = mapping
        self.primary = primary

    def __repr__(self):
        if self.primary:
            return "<IndexDetails key={} unique={} primary>".format(
                self.key, self.unique
            )
        return "<IndexDetails key={} unique={}>".format(self.key, self.unique)

    def get_values(self, model):
        return self.values_func(self.attribute.read(model))



"""Types are used to define the internal storage in ZODB.

There are no need to setup specific types in basic usage of sheraf
because `Model` uses `xAttribute`, not `Type`.
"""


import BTrees.Length
import BTrees.OOBTree
import persistent
import ZODB

import sheraf.tools.dicttools

from .largedict import LargeDict
from .largelist import LargeList

assert LargeDict
assert LargeList


SmallList = persistent.list.PersistentList
List = SmallList
Set = BTrees.OOBTree.OOTreeSet


class SmallDict(persistent.mapping.PersistentMapping):
    """SmallDict is a :class:`PersistentMapping` implementation with a simple
    conflict resolution implementation.

    When two different keys of the mapping are edited in concurrency, no
    conflict is raised.
    """

    def _p_resolveConflict(self, old, saved, commited):
        try:
            return sheraf.tools.dicttools.merge(old, saved, commited)
        except sheraf.tools.dicttools.DictConflictException:
            raise ZODB.POSException.ConflictError()



class SherafException(Exception):
    """Main sheraf exception."""

    pass


class ObjectNotFoundException(SherafException):
    """Raised when trying to read an unexisting object."""

    pass


class ModelObjectNotFoundException(ObjectNotFoundException):
    """Raised when trying to read an unexisting :class:`ModelObject`.

    :param model_class: Model class of the unfound object
    :type model_class: sheraf.models.BaseModelMetaclass
    :param object_id: Id of the unfound object
    :type object_id: str
    """

    MESSAGE_TPL = "Key {identifier} not found in {model_name}, '{index_name}' index"

    def __init__(self, model_class, identifier, index_name=None):
        self.model_class = model_class
        self.identifier = identifier
        index_name = index_name or model_class.primary_key()

        message = self.MESSAGE_TPL.format(
            identifier=repr(identifier),
            model_name=self.model_class.__name__,
            index_name=index_name,
        )

        super().__init__(message)


class SameNameForTableException(SherafException):
    """Raised when two models have the same "table" attribute.

    >>> class FirstModel(sheraf.Model):
    ...     table = "first_model"
    ...
    >>> class SecondModel(sheraf.Model):
    ...     table = "first_model"
    Traceback (most recent call last):
        ...
    sheraf.exceptions.SameNameForTableException: Table named 'first_model' used twice (FirstModel and SecondModel)
    """


class ConnectionAlreadyOpened(SherafException):
    """Raised when user tries to open an connection when a connection is
    already opened.

    >>> with sheraf.connection():
    ...     with sheraf.connection():
    ...         do_amazing_stuff()
    Traceback (most recent call last):
        ...
    sheraf.exceptions.ConnectionAlreadyOpened: First connection was <Connection at ...> on ... at line ...
    """


class NotConnectedException(SherafException):
    """Raised when trying to handle things in the database while not being
    connected.

    >>> with sheraf.connection():
    ...     sheraf.attempt(do_amazing_stuff)  # produces amazing stuff
    ...
    >>> sheraf.attempt(do_amazing_stuff)
    Traceback (most recent call last):
        ...
    sheraf.exceptions.NotConnectedException
    """


class InvalidFilterException(SherafException):
    """Raised when an invalid :class:`~sheraf.queryset.QuerySet` filter has been
    called.

    >>> class MyModel(sheraf.Model):
    ...    my_attribute = sheraf.SimpleAttribute()
    ...
    >>> with sheraf.connection():
    ...    MyModel.filter(foobar=True)
    Traceback (most recent call last):
        ...
    sheraf.exceptions.InvalidFilterException: MyModel has no attribute foobar
    """


class InvalidOrderException(SherafException):
    """Raised when an invalid :class:`~sheraf.queryset.QuerySet` order has been
    called.

    >>> class MyModel(sheraf.Model):
    ...    table = "my_model"
    ...    my_attribute = sheraf.SimpleAttribute()
    ...
    >>> with sheraf.connection():
    ...    MyModel.all().order(foobar=True)
    Traceback (most recent call last):
        ...
    sheraf.exceptions.InvalidOrderException: MyModel has no attribute foobar
    >>> with sheraf.connection():
    ...    MyModel.all().order(my_attribute=sheraf.ASC) \\
    ...                 .order(my_attribute=sheraf.DESC)  # Raises an InvalidOrderException
    Traceback (most recent call last):
        ...
    sheraf.exceptions.InvalidOrderException: Some order parameters appeared twice
    """


class QuerySetUnpackException(SherafException):
    """Raised when calling :func:`~sheraf.queryset.QuerySet.get` on
    :class:`~sheraf.queryset.QuerySet` containing invalid number of elements.

    >>> class Cowboy(sheraf.Model):
    ...     table = "people"
    ...     name = sheraf.SimpleAttribute()
    ...
    >>> with sheraf.connection():
    ...     peter = Cowboy.create(name="Peter")
    ...
    ...     assert peter == Cowboy.filter(name="Peter").get()
    ...
    >>> # Fails with too many values
    ... with sheraf.connection():
    ...     Cowboy.create(name="Peter")
    ...     Cowboy.create(name="Steven")
    ...     Cowboy.all().get()
    Traceback (most recent call last):
        ...
    sheraf.exceptions.QuerySetUnpackException: Trying to unpack more than 1 value from a QuerySet
    """


class EmptyQuerySetUnpackException(QuerySetUnpackException):
    """Raised when calling :func:`~sheraf.queryset.QuerySet.get` on
    :class:`~sheraf.queryset.QuerySet` containing 0 element.

    >>> # Fails with zero value
    ... with sheraf.connection():
    ...     Cowboy.filter(name="unknown cowboy").get()
    Traceback (most recent call last):
    ...
    sheraf.exceptions.EmptyQuerySetUnpackException: Trying to unpack an empty QuerySet
    """


class InvalidIndexException(SherafException):
    """
    This exception is raised by :func:`~sheraf.models.BaseModel.read` and
    :func:`~sheraf.models.BaseModel.read_these` when a parameter is passed
    and is not a valid index.

    >>> class Cowboy(sheraf.Model):
    ...     table = "cowboy"
    ...     age = sheraf.IntegerAttribute().index()
    ...
    >>> with sheraf.connection():
    ...     Cowboy.read(size=4)
    Traceback (most recent call last):
    ...
    sheraf.exceptions.InvalidIndexException: 'size' is not a valid index
    """


class UniqueIndexException(InvalidIndexException):
    """
    This exception is raised when a value is set twice in an unique index.

    >>> class Cowboy(sheraf.Model):
    ...     table = "cowboy"
    ...     name = sheraf.StringAttribute().index(unique=True)
    ...
    >>> with sheraf.connection():
    ...     Cowboy.create(name="George Abitbol")
    ...     Cowboy.create(name="George Abitbol")
    Traceback (most recent call last):
    ...
    sheraf.exceptions.UniqueIndexException: The key 'George Abitbol' is already present in the index 'name'
    """


class MultipleIndexException(InvalidIndexException):
    """
    This exception is raised by :func:`~sheraf.models.BaseModel.read` when a
    multiple index is passed as a positionnal argument.

    >>> class Cowboy(sheraf.Model):
    ...     table = "cowboy"
    ...     name = sheraf.StringAttribute()
    ...     hobby = sheraf.StringAttribute().index()
    ...
    >>> with sheraf.connection():
    ...     Cowboy.create(name="George", hobby="nice hats")
    ...     Cowboy.create(name="Peter", hobby="nice hats")
    ...     Cowboy.read(hobby="nice hats")
    Traceback (most recent call last):
    ...
    sheraf.exceptions.MultipleIndexException: 'hobby' is a multiple index and cannot be used with 'read'
    """


class IndexationWarning(UserWarning):
    """
    This warning is emitted when you edit or create a model instance which has an outdated
    indexation table.
    """


class PrimaryKeyException(SherafException):
    """
    This exception is raised when issues happens with index primary keys.

    When creating a model with zero, or several primary indexes.

    >>> class Horse(sheraf.AttributeModel):
    ...     name = sheraf.StringAttribute().index(primary=True)
    ...     breed = sheraf.StringAttribute().index(primary=True)
    ...
    >>> class Cowboy(sheraf.Model):
    ...     name = sheraf.StringAttribute()
    ...     horses = sheraf.IndexedModelAttribute(Horse)
    ...
    >>> with sheraf.connection():
    ...     george = Cowboy.create(name="George Abitbol")
    ...     george.horses.create(name="Jolly Jumper", breed="shetland")
    Traceback (most recent call last):
    ...
    sheraf.exceptions.PrimaryKeyException: "A model can have only one primary key. 'Horse' has 'name' and 'breed'"
    """



from itertools import filterfalse


def unique_everseen(iterable, key=None):
    """
    List unique elements, preserving order. Remember all elements ever seen.
    >>> from sheraf.tools.more_itertools import unique_everseen
    >>> list(unique_everseen('AAAABBBCCDAABBB'))
    ['A', 'B', 'C', 'D']
    >>> list(unique_everseen('ABBCcAD', str.lower))
    ['A', 'B', 'C', 'D']
    """
    seen = set()
    seen_add = seen.add
    if key is None:
        for element in filterfalse(seen.__contains__, iterable):
            seen_add(element)
            yield element
    else:
        for element in iterable:
            k = key(element)
            if k not in seen:
                seen_add(k)
                yield element



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['sheraf',
 'sheraf.attributes',
 'sheraf.batches',
 'sheraf.models',
 'sheraf.tools',
 'sheraf.types']

package_data = \
{'': ['*']}

install_requires = \
['orderedset', 'zodb>=5', 'zodburi']

extras_require = \
{'all': ['zeo', 'psycopg2', 'psycopg2cffi', 'relstorage', 'colored', 'tqdm'],
 'doc': ['recommonmark', 'sphinx', 'sphinx-rtd-theme', 'sphinx-issues'],
 'relstorage_pg': ['psycopg2', 'psycopg2cffi', 'relstorage'],
 'zeo': ['zeo']}

setup_kwargs = {
    'name': 'sheraf',
    'version': '0.2.1',
    'description': 'Versatile ZODB abstraction layer',
    'long_description': '# A versatile ZODB abstraction layer\n\nsheraf is an overlay over ZODB. It provides a high level interface to save and read your data.\n\n## Installation\n\nsheraf is compatible with Python 3.5+\n\n    poetry add sheraf\n    # or\n    pip install sheraf\n\nIf you need pytest fixtures for your project check out [pytest-sheraf](https://gitlab.com/yaal/pytest-sheraf). There are also [sheraf fixtures for unittest](https://gitlab.com/yaal/unittest-sheraf).\n\n    pip install pytest-sheraf\n\nThe documentation is [hosted here](https://sheraf.readthedocs.io/en/latest/).\n\n## Contributing\n\nBug reports and pull requests are highly encouraged!\n\n - Test some code : `poetry run pytest` and `poetry run tox`\n - Format code :\xa0`black`\n - Generate documentation : `poetry run tox -e doc`\n\n## Documentation\n\n    poetry run tox -e doc\n    open build/sphinx/html/index.html\n\n## Development installation\n\nsheraf use poetry as its main build tool. Do not hesitate to check [the documentation](https://python-poetry.org/docs/).\n\n    poetry install --extras all\n',
    'author': 'Yaal team',
    'author_email': 'contact@yaal.fr',
    'maintainer': 'Éloi Rivard',
    'maintainer_email': 'eloi@yaal.fr',
    'url': 'https://sheraf.readthedocs.io/en/latest/',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.5,<4',
}


setup(**setup_kwargs)



from BTrees.OOBTree import OOBTree
from sheraf.attributes.indexdetails import IndexDetails


READ_MEMOIZATION = False
WRITE_MEMOIZATION = True


def set_read_memoization(should_memoize_read):
    global READ_MEMOIZATION
    READ_MEMOIZATION = should_memoize_read


def set_write_memoization(should_memoize_write):
    global WRITE_MEMOIZATION
    WRITE_MEMOIZATION = should_memoize_write


class BaseAttribute(object):
    """
    Base type of all attributes of a base model.

    :param default: The value this attribute will be initialized with. If it is a callable object,
                    it will be called on initialization, else it will simply be copied. The callable
                    object can take either no argument, or one argument that will be the parent model.
    :type default: a callable object or a simple object
    :param key: The key to identify the attribute in its parent persistent mapping.
    :param lazy: If True, the objet carried by the attribute is created on the first
                          read or write access. If False, it is created when the model object
                          is created. Default is True.
    :type lazy: :class:`bool`
    :param read_memoization: Whether this attribute should be memoized on read. ``False`` by default.
    :type read_memoization: :class:`bool`
    :param write_memoization: Whether this attribute should be memoized on write. ``True`` by default.
    :type write_memoization: :class:`bool`

    When an attribute is memoized, its next reading will not result in a new database access.
    Attributes:
    - indexes:    a dictionary of Indexes. The key with value None stands for this attribute's name.
    """

    default_index_mapping = OOBTree

    def __init__(
        self,
        default=None,
        key=None,
        lazy=True,
        read_memoization=None,
        write_memoization=None,
        store_default_value=True,
    ):
        self._default_value = default
        self._default_key = None
        self._key = key
        self.read_memoization = (
            READ_MEMOIZATION if read_memoization is None else read_memoization
        )
        self.write_memoization = (
            WRITE_MEMOIZATION if write_memoization is None else write_memoization
        )
        self.lazy = lazy
        self.store_default_value = store_default_value
        self.indexes = {}

    def set_default_key(self, key):
        self._default_key = key

    def create(self, parent):
        """
        :return: The attribute's defined default value. If it is a callable, return
                 the result of calling it instead.
        """
        if not callable(self._default_value):
            return self._default_value

        try:
            return self._default_value()
        except TypeError:
            return self._default_value(parent)

    def is_created(self, parent):
        """
        :param parent: The owner of this attribute
        :return: true if the object is effectively created"""
        return self.key(parent) in parent.mapping

    def key(self, parent):
        """:return: the key that identifies this attribute in its owner object"""
        if self._key is None:
            return self._default_key

        if isinstance(self._key, (list, tuple)):
            for key in self._key:
                if key in parent.mapping:
                    return key
            return self._key[0]

        return self._key

    def read_raw(self, parent):
        # Internal.
        # :param parent: The owner of this attribute
        # :return: the raw representation of this attribute (ie as stored in ZODB)
        try:
            return parent.mapping[self.key(parent)]
        except KeyError:
            default_value = self.serialize(self.create(parent))
            if self.store_default_value:
                parent.mapping[self.key(parent)] = default_value
            return default_value

    def write_raw(self, parent, value):
        # Internal.
        # :param parent: The owner of this attribute
        # :param value: The value assigned to this attribute
        # assigns the value parameter to this attribute
        parent.mapping[self.key(parent)] = value
        return value

    def serialize(self, value):
        """Get data and transform it into something ZODB can store.

        :param value: The input data.
        :return: The transformed data.
        """
        return value

    def deserialize(self, value):
        """Get raw data and transform it into something more convenient to use.

        :param value: The input data.
        :return: The transformed data.
        """
        return value

    def read(self, parent):
        """Reads some raw data from the parent model and transform it into
        something more convenient to use. Most of the time, you should use
        :func:`~sheraf.attributes.base.BaseAttribute.deserialize`.

        :param parent: The model parent.
        :return: the transformed value of this attribute
        """
        return self.deserialize(self.read_raw(parent))

    def write(self, parent, value):
        """Takes user data, transform it into something ZODB can store, and
        store it into the model parent persistent.

        :param parent: The owner object of this attribute
        :param value: The user data to store.
        :return: The stored and reread data. In some cases it may be different that the input data.
        """
        written_value = self.write_raw(parent, self.serialize(value))
        return self.deserialize(written_value)

    def update(
        self,
        old_value,
        new_value,
        addition=True,
        edition=True,
        deletion=False,
        replacement=False,
    ):
        """Updates the value of the attribute.

        :param old_value: The previous value the attribute had.
        :param new_value: The new value that the attribute should be updated with.
        :param addition: On collections, adds elements to the attribute if they are present in `new_value` and not in `old_value`, defaults to `True`.
        :param edition: On collections, edits the attribute element if present on both `old_value` and `new_value`, defaults to `True`.
        :param deletion: On collections, removes an element if present in `old_value` and absent from `new_value`, default to `False`.
        :param replacement: Replace sub-collections or sub-models instead of editing them in-place, defaults to `False`.
        """
        return new_value if edition else old_value

    def delattr(self, parent):
        # Internal. delete attribute from its owner
        # TODO: We should remove this try/except
        try:
            del parent.mapping[self.key(parent)]
        except KeyError:
            pass

    def save(self, parent):
        pass

    def delete(self, parent):
        pass

    def index(
        self,
        unique=False,
        key=None,
        values=None,
        search=None,
        mapping=None,
        primary=False,
    ):
        """
        Indexing an attribute allows very fast reading with :func:`~sheraf.queryset.QuerySet.filter` calls.

        :param unique: If the attribute is unique, and two models have the same value for this
                      attribute, a :class:`~sheraf.exceptions.UniqueIndexException` is raised
                      when trying to write the second one. Automatically set to :class:`True` if
                      *primary* is :class:`True`.
        :param key: The key the index will use. By default, just the attribute name is used.
        :param values: A callable that takes the current attribute value and returns a collection of values to index. Each generated value will be indexed each time this attribute is edited. It may take time if the generated collection is large. By default, the current attribute raw value is used.
        :param primary: If true, this will be the default index for the model. `False` by default.

        When indexes are used, **lazy** is disabled.

        >>> class People(sheraf.Model):
        ...     table = "index_people"
        ...
        ...     # Simple indexing
        ...     name = sheraf.SimpleAttribute().index()
        ...
        ...     # Emails can only be owned once
        ...     email = sheraf.SimpleAttribute().index(unique=True)
        ...
        ...     # Indexing people by their decade
        ...     age = sheraf.SimpleAttribute().index(key="decade", values=lambda age: {age // 10})
        ...
        >>> with sheraf.connection(commit=True):
        ...     m = People.create(name="George Abitbol", email="george@abitbol.com", age=55)
        ...
        >>> with sheraf.connection():
        ...     assert [m] == People.filter(name="George Abitbol")
        ...     assert [m] == People.filter(decade=5)
        ...
        >>> with sheraf.connection():
        ...     People.create(name="Peter", email="george@abitbol.com", age=35)
        Traceback (most recent call last):
            ...
        UniqueIndexException
        """
        self.indexes[key] = IndexDetails(
            self,
            unique,
            key,
            values,
            search,
            mapping or self.default_index_mapping,
            primary,
        )
        self.lazy = False

        return self
